#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include "softcris.h"
#include "hdfi.h"
#include "crisExpEvt.h"
#include "ce_crisEvt.h"
#include "level1.h"
#include "cris/crisGetData/getEvents_c.h"
#include "mewRecompressedEvent_c.h"

    /* debug */
#define VERBOSE 0
#define VERBOSE_SOFT_EXCEPTIONS_FOR_REJECTED_EVENTS 0

    /* assume there won't be more than 500 events in an instrument cycle */
#define MAX_EVENTS_PER_FRAME 500

#define COMPRESSED_ELEMENTS 256
#define NUMBER_OF_DESIRED_TRAJECTORIES 1

u_char compressNpix(long npix);
long uncompNpix(u_char compNpix);
double interpol(u_char *compressed, long *uncompressed, long npix);
u_char compressInt(long intensity);
void processSOFTException(int errorCode);
 
struct mewRecompressedEvent_c *recompressFramesEvents(struct
					ce_crisExpEvt *pExpEvt, int eventCount);
void writeFramesRecompressedEvents(int fd,
	struct mewRecompressedEvent_c *pRecompressedEvents, int eventCount);
 
int main(int argc, char *argv[]) {
    int day, startDay1996, endDay1996;
    char defaultEventsDirectory[] = ".";
    char *eventsDirectory;

    if ((argc < 3) || (argc > 4)) {
	printf("Usage: getCrisMEWFormatEvents ");
	printf("startDayOf1996 endDayOf1996 [eventsDir]\n");
	return 0;
    }

    startDay1996 = atoi(argv[1]);
    endDay1996 = atoi(argv[2]);
    if (argc == 4) eventsDirectory = (char *) argv[3];
    else eventsDirectory = defaultEventsDirectory;

	/* set trajectory method in SOFT lib */
	/* brightest-each-layer -> 0, fiber-fit -> 1-4, multi-traj -> 5-8 */
    sc_traj_select_method(8);

	/* Begin looping through event cycles on a given day. */
    for (day = startDay1996; day <= endDay1996; day++) {
	int fd;
	char filePath[256];
	uint32 *pRequestTimes;

		/* check existance of inputArchiveFile */
	{
	    static char pathBuffer[256];
	    static const char instrumentID = 'c';
	    static const char dataFileSuffix[] = "evt";
	    struct tm *pTimeStruct;
	    
	    pTimeStruct = getDateFromDay1996(day);
	    generateDataFilePath(pathBuffer, pTimeStruct, instrumentID,
								dataFileSuffix);

	    if (!fileExists(pathBuffer)) {	/* go to next day */
		fprintf(stderr, "Warning: Input archive file %s ", pathBuffer);
		fprintf(stderr, "does not exist.\n");
		fprintf(stderr, "\tContinuing to next day.\n\n");
		continue;
	    }
	}

	/* sprintf(filePath, "%s/events.%4.4u", eventsDirectory, day); */
	sprintf(filePath, "%s/c_%4.4u.ce1", eventsDirectory, day);
	fd = openOrCreateFileToReadWrite(filePath);

	pRequestTimes = getRequestTimesForDaysEvents_c(day, 1);
	while (*pRequestTimes != 0) {
	    static struct ce_crisEvt *pL1Event;
	    static struct ce_crisExpEvt pExpEvt[MAX_EVENTS_PER_FRAME];
	    u_char goodTrajectory;
	    int iEvt, j;

	    iEvt = 0;
	    pL1Event = getEvents_c(*pRequestTimes);

	    while (pL1Event->secs1996 != 0) {
		struct sc_Coords_Traj
		    coordsAndTrajectoriesInfo[NUMBER_OF_DESIRED_TRAJECTORIES];
		int errorCode, numberOfTrajectories;
		static struct sc_fmapElem *pFiberMapElement;
		int result, i;

		if (iEvt >= MAX_EVENTS_PER_FRAME) {
		    printf("Error: Exceeding maximum of %d events/frame.\n",
							MAX_EVENTS_PER_FRAME);
		    printf("Exitting.");
		    exit(1);
		}
	    
		goodTrajectory = 0;   /* assume bad trajectory at start */
		for (i = 0; i < CE_CRIS_MAXBLOBS; i++) { /* initialize */
		    pExpEvt[iEvt].blobpos[i] = -100.0; /* needs to be unique */
		    pExpEvt[iEvt].layer[i] = (u_char) 32;
		}
		pExpEvt[iEvt].x1 = 0.0;
		pExpEvt[iEvt].x2 = 0.0;
		pExpEvt[iEvt].x3 = 0.0;
		pExpEvt[iEvt].y1 = 0.0;
		pExpEvt[iEvt].y2 = 0.0;
		pExpEvt[iEvt].y3 = 0.0;
		pExpEvt[iEvt].ix1 = (u_char) 0;	/* not used */
		pExpEvt[iEvt].ix2 = (u_char) 0;	/* not used */
		pExpEvt[iEvt].ix3 = (u_char) 0;	/* not used */
		pExpEvt[iEvt].iy1 = (u_char) 0;	/* not used */
		pExpEvt[iEvt].iy2 = (u_char) 0;	/* not used */
		pExpEvt[iEvt].iy3 = (u_char) 0;	/* not used */
		pExpEvt[iEvt].nx1 = (u_char) 0;
		pExpEvt[iEvt].nx2 = (u_char) 0;
		pExpEvt[iEvt].nx3 = (u_char) 0;
		pExpEvt[iEvt].ny1 = (u_char) 0;
		pExpEvt[iEvt].ny2 = (u_char) 0;
		pExpEvt[iEvt].ny3 = (u_char) 0;
		pExpEvt[iEvt].goodtraj = (u_char) 0;

		for (i = 0; i < ((int) pL1Event->nblobs); i++) {
		    float xreal, yreal;
			/* layer ID and middle fiber ID for each layer */
		    static int layerAndMidFiberIDs[6] =
					{ 1940, 640, 4503, 3248, 6828, 5686 };
		    int layer;

		    pFiberMapElement = sc_centr_to_fiber(SC_CAM2,
					(double) pL1Event->hodo[i].x,
					(double) pL1Event->hodo[i].y);
			/* unmapped coords? */
		    if ((pFiberMapElement == NULL) ||
					(pFiberMapElement->layer == -1)) {
			pExpEvt[iEvt].blobpos[i] = -10.0;
			pExpEvt[iEvt].layer[i] = (u_char) 255;
			continue;
		    }

		    layer = 3 - ((int) (pFiberMapElement->layer / 2));

		    if ((pFiberMapElement->layer % 2) == 0) {
			    /* this is a y-layer fiber */
			result = sc_fiber_to_real(SC_CAM2, layer,
				layerAndMidFiberIDs[pFiberMapElement->layer],
				pFiberMapElement->fiber, &xreal, &yreal);
			if (result == 0) { /* all OK */
			    pExpEvt[iEvt].blobpos[i] = (float) yreal;
			    pExpEvt[iEvt].layer[i] = pFiberMapElement->layer;
			}
			else {
			    pExpEvt[iEvt].blobpos[i] = -1.0;
			    pExpEvt[iEvt].layer[i] = pFiberMapElement->layer;
			}
		    }
		    else {			/* this is an x-layer fiber */
			result = sc_fiber_to_real(SC_CAM2, layer,
				pFiberMapElement->fiber,
				layerAndMidFiberIDs[pFiberMapElement->layer],
				&xreal, &yreal);
			if (result == 0) {
			    pExpEvt[iEvt].blobpos[i] = (float) xreal;
			    pExpEvt[iEvt].layer[i] = pFiberMapElement->layer;
			}
			else {
			    pExpEvt[iEvt].blobpos[i] = -1.0;
			    pExpEvt[iEvt].layer[i] = pFiberMapElement->layer;
			}
		    } 
		}

		numberOfTrajectories = NUMBER_OF_DESIRED_TRAJECTORIES;
		result = sc_cris_traj(pL1Event, coordsAndTrajectoriesInfo,
							&numberOfTrajectories);
		errorCode = sc_get_err_code();

		if (result == 0) {		/* found a good trajectory */
		    int blobID, blobsFoundCount, blobFound[6];
		    short x, y;

			/* this flag is being set to 2 instead of 1 in order */
			/* that the generated data files agree with an older */
			/* version of this routine. */
		    goodTrajectory = 2;

			/* top layer to bottom layer in SOFT */
		    pExpEvt[iEvt].x1 = coordsAndTrajectoriesInfo[0].T.layer1x;
		    pExpEvt[iEvt].y1 = coordsAndTrajectoriesInfo[0].T.layer1y;
		    pExpEvt[iEvt].x2 = coordsAndTrajectoriesInfo[0].T.layer2x;
		    pExpEvt[iEvt].y2 = coordsAndTrajectoriesInfo[0].T.layer2y;
		    pExpEvt[iEvt].x3 = coordsAndTrajectoriesInfo[0].T.layer3x;
		    pExpEvt[iEvt].y3 = coordsAndTrajectoriesInfo[0].T.layer3y;

			/* here is the quality factor of the trajectory */
		    /* coordsAndTrajectoriesInfo[0].T.Q; */

			/* determine th blob IDs used in the trajectory.
		         *     here are the specs given regarding layers:
			 *     (t == coordsAndTrajectoriesInfo)
			 *
			 *  x coordinate   y coordinate   intensity    of blob
			 * t[0].C.xpix[0] t[0].C.ypix[0] t[0].C.I[0]  H3 Y layer
			 * t[0].C.xpix[1] t[0].C.ypix[1] t[0].C.I[1]  H3 X layer
			 * t[0].C.xpix[2] t[0].C.ypix[2] t[0].C.I[2]  H2 Y layer
			 * t[0].C.xpix[3] t[0].C.ypix[3] t[0].C.I[3]  H2 X layer
			 * t[0].C.xpix[4] t[0].C.ypix[4] t[0].C.I[4]  H1 Y layer
			 * t[0].C.xpix[5] t[0].C.ypix[5] t[0].C.I[5]  H1 X layer
			 *
			 */

			/* go through the blobs in the raw data and look for
			 * matches with the blobs returned with the trajectory
			 * in order to determine their blobIDs.
			 */

			/* flags for blob found in each layer */
		    for (j = 0; j < 6; j++) blobFound[j] = 0;

		    blobsFoundCount = 0;
		    blobID = 0;
		    while (blobsFoundCount < 6) {
			    /* x and y pixel coords corresponding to blobID */
			x = pL1Event->hodo[blobID].x;
			y = pL1Event->hodo[blobID].y;

			    /* j = 0 is H3 Y layer, j = 1 is H3 X layer, etc. */
			for (j = 0; j < 6; j++) {
			    if (!blobFound[j] &&
			      (coordsAndTrajectoriesInfo[0].C.xpix[j] == x) &&
			      (coordsAndTrajectoriesInfo[0].C.ypix[j] == y)) {
				blobFound[j] = 1;
				blobsFoundCount++;
				switch (j) {
				    case 0: /* blob used in the H3 Y layer */
					pExpEvt[iEvt].ny3 = blobID;
					break;
				    case 1: /* blob used in the H3 X layer */
					pExpEvt[iEvt].nx3 = blobID;
					break;
				    case 2: /* blob used in the H2 Y layer */
					pExpEvt[iEvt].ny2 = blobID;
					break;
				    case 3: /* blob used in the H2 X layer */
					pExpEvt[iEvt].nx2 = blobID;
					break;
				    case 4: /* blob used in the H1 Y layer */
					pExpEvt[iEvt].ny1 = blobID;
					break;
				    case 5: /* blob used in the H1 X layer */
					pExpEvt[iEvt].nx1 = blobID;
					break;
				}
				break;
			    }
			}

			    /* go to next blob in the raw data */
			blobID++;
			    /* blobID ranges from 0 to CE_CRIS_MAXBLOBS - 1 */
			if (blobID >= CE_CRIS_MAXBLOBS) {
			    if (blobsFoundCount == 6) break;
			    else if (VERBOSE && (blobsFoundCount == 5))
				fprintf(stderr, "Only 5 Blobs in trajectory\n");
			    else if (VERBOSE && (blobsFoundCount == 4))
				fprintf(stderr, "Only 4 Blobs in trajectory\n");
			    else if (blobsFoundCount < 4) {
				fprintf(stderr, "ERROR: 3 or more blobs in ");
				fprintf(stderr, "trajectory not found!\n");
				fprintf(stderr,"(This should never happen.)\n");
			    }

			    if (VERBOSE) {
				fprintf(stderr, "\tevent time stamp: %ld\n",
							pL1Event->secs1996);
				fprintf(stderr, "\tevent number: %ld\n",
							pL1Event->evtno);
				fprintf(stderr, "\toffset: %u\n",
							pL1Event->evt_offset);
				fprintf(stderr, "\t%d %d %d %d %d %d\n",
				    blobFound[0], blobFound[1], blobFound[2],
				    blobFound[3], blobFound[4], blobFound[5]);
			    }

				/* assign dummy id to unfound blobs */
			    if (!blobFound[5]) pExpEvt[iEvt].nx1 = 0xff;
			    if (!blobFound[4]) pExpEvt[iEvt].ny1 = 0xff;
			    if (!blobFound[3]) pExpEvt[iEvt].nx2 = 0xff;
			    if (!blobFound[2]) pExpEvt[iEvt].ny2 = 0xff;
			    if (!blobFound[1]) pExpEvt[iEvt].nx3 = 0xff;
			    if (!blobFound[0]) pExpEvt[iEvt].ny3 = 0xff;

			    break;
			}
		    }
		}
		else processSOFTException(errorCode);

		pExpEvt[iEvt].EvtData = *pL1Event;
		pExpEvt[iEvt].goodtraj = goodTrajectory;

		pL1Event++;
		iEvt++; 
	    }

	    /* printf("frame= %d, eventCount = %d\n", loop, iEvt); */

		/* recompress events and write them to file */
	    {
		struct mewRecompressedEvent_c *pRecompressedEvents,
						*pRecompressedEventsStart;
		pRecompressedEventsStart = recompressFramesEvents(pExpEvt,iEvt);
		pRecompressedEvents = pRecompressedEventsStart;
		writeFramesRecompressedEvents(fd, pRecompressedEvents, iEvt);
		free(pRecompressedEventsStart);
	    }

	    pRequestTimes++;
	}
	close(fd);
    }
    return 0;
}

void writeFramesRecompressedEvents(int fd,
		struct mewRecompressedEvent_c *pRecompressedEvents, int count) {
    long bytesLeftToWrite;
    int DEVICE_BLOCK_SIZE = 8192;
    char *pChar;

    bytesLeftToWrite = sizeof(struct mewRecompressedEvent_c) * count;
    pChar = (char *) pRecompressedEvents;
    while (bytesLeftToWrite) {
	if (bytesLeftToWrite >= DEVICE_BLOCK_SIZE) {
	    writeToFile(fd, pChar, DEVICE_BLOCK_SIZE);
	    bytesLeftToWrite -= DEVICE_BLOCK_SIZE;
	    pChar += DEVICE_BLOCK_SIZE;
	}
	else {
	    writeToFile(fd, pChar, (int) bytesLeftToWrite);
	    break;
	}
    }
}

struct mewRecompressedEvent_c *recompressFramesEvents(
				struct ce_crisExpEvt *pExpEvt, int eventCount) {
    int i, j, iStack, iGuard, iBlob, iLayer, iXY;
    int CENTER_THRESH[] = { 125, 450, 1000 };
    int GUARD_THRESH[] =  { 125, 450, 1000 };
    short check;

    struct mewRecompressedEvent_c *pRecompressedEventsStart;
    struct mewRecompressedEvent_c *pRecompressedEvents;

    pRecompressedEventsStart = (struct mewRecompressedEvent_c *)
		calloc(eventCount, sizeof(struct mewRecompressedEvent_c));
    pRecompressedEvents = pRecompressedEventsStart;

    for (i = 0; i < eventCount; i++) {
	pRecompressedEvents->sec96 = pExpEvt->EvtData.secs1996;
	pRecompressedEvents->sc_clk = pExpEvt->EvtData.evtno;
	pRecompressedEvents->ev_ofs = pExpEvt->EvtData.evt_offset;
	pRecompressedEvents->ev_len = pExpEvt->EvtData.nbytes;

	if (pExpEvt->EvtData.prim_ctr_ph[1] > pExpEvt->EvtData.prim_ctr_ph[0])
	    pRecompressedEvents->flags = 0x01;
	else pRecompressedEvents->flags = 0x00;
	pRecompressedEvents->flags = pRecompressedEvents->flags |
			((pExpEvt->EvtData.prim_tel_pair << 1) & 0x02) |
			((pExpEvt->EvtData.guard_hi_or   << 2) & 0x04) |
			((pExpEvt->EvtData.got_guard     << 3) & 0x08) |
			((pExpEvt->EvtData.sec_tel_act   << 4) & 0x30) |
			((pExpEvt->EvtData.got_2nd_tel   << 6) & 0x40) |
			((pExpEvt->EvtData.hazard        << 7) & 0x80);

	pRecompressedEvents->bufid = (pExpEvt->EvtData.buffer_id & 0x3F);

	pRecompressedEvents->rng_gd = ((pExpEvt->EvtData.R1 & 0x0F) |
					((pExpEvt->EvtData.G1 << 4) & 0x70));
	pRecompressedEvents->rng_gd2 = ((pExpEvt->EvtData.R2 & 0x0F) |
					((pExpEvt->EvtData.G2 << 4) & 0x70));

	for (iStack = 0; iStack < 10; iStack++) {
	    short centerThreshCmp = 0;
	    for (j = 0; j < 3; j++)
		if (pExpEvt->EvtData.sec_ctr_ph[iStack] >= CENTER_THRESH[j])
			centerThreshCmp++;
	    pRecompressedEvents->stk[iStack] =
		(pExpEvt->EvtData.prim_ctr_ph[iStack] & 0x0FFF) |
		((pExpEvt->EvtData.prim_tag << (12 - iStack)) & 0x1000) |
		((centerThreshCmp << 13) & 0x6000) |
		((pExpEvt->EvtData.sec_tag << (15 - iStack)) & 0x8000);
	}

	for (iGuard = 0; iGuard < 6; iGuard++) {
	    u_short guardThreshCmp = 0, primTagAux;
	    for (j = 0; j < 3; j++) 
		if (pExpEvt->EvtData.sec_guard_ph[iGuard] >= GUARD_THRESH[j])
			guardThreshCmp++;
	    primTagAux = pExpEvt->EvtData.prim_tag;
	    if (iGuard > 2) primTagAux = primTagAux >> (iGuard - 2);
	    else primTagAux = primTagAux << (2 - iGuard);
	    pRecompressedEvents->grd[iGuard] =
		(pExpEvt->EvtData.prim_guard_ph[iGuard] & 0x0FFF) |
		(primTagAux & 0x1000) |
		((guardThreshCmp << 13) & 0x6000) |
		((pExpEvt->EvtData.sec_tag << (5 - iGuard)) & 0x8000);
	}

	if ( (pExpEvt->EvtData.stim != 0x00) &
		((pExpEvt->EvtData.stimdata.ctr_dac_level !=
				pExpEvt->EvtData.stimdata.guard_dac_level) |
		(pExpEvt->EvtData.stimdata.ctr_pulsers != 0x0F) |
		(pExpEvt->EvtData.stimdata.guard_pulsers != 0x0F) |
		(pExpEvt->EvtData.stimdata.ctr_hi_gain != 0x01) |
		(pExpEvt->EvtData.stimdata.guard_hi_gain != 0x01)) )
	    check = 0x4000;
	else check = 0x0000;
	pRecompressedEvents->stim =
		((pExpEvt->EvtData.stimdata.ctr_dac_level & 0xFF) |
		((pExpEvt->EvtData.stimdata.soft_led_pulse << 8) & 0x3F00) |
		check |
		((pExpEvt->EvtData.stim << 15) & 0x8000));

	pRecompressedEvents->npix = compressNpix(pExpEvt->EvtData.npix);

	pRecompressedEvents->soft_par =
		(pExpEvt->EvtData.nblobs & 0x1F) |
		((pExpEvt->EvtData.cam << 6) & 0x40) |
		((pExpEvt->EvtData.centroid_err << 7) & 0x80);

	    /* set all the status bytes to 0.  not yet defined. */
	for (j = 0; j < 4; j++) pRecompressedEvents->status[j] = 0x00;

	for (iBlob = 0; iBlob < 18; iBlob++) {
	    if (iBlob < pExpEvt->EvtData.nblobs) {
		u_char min;
		int position;

		if (pExpEvt->layer[iBlob] < 0x06) min = pExpEvt->layer[iBlob];
		else min = 0x06;
		pRecompressedEvents->soft_dat[iBlob][0] = (u_char)
			((compressInt(pExpEvt->EvtData.hodo[iBlob].I) & 0x1F) |
			((min << 5) & 0xE0));

		    /* convert blob position (mm) to integer in 50 um */
		    /* units.  round to the nearest value.            */
		position = (int) ((pExpEvt->blobpos[iBlob] / 0.050) + 0.5);
		if (pExpEvt->layer[iBlob] <= 5)
		    pRecompressedEvents->soft_dat[iBlob][1] =
						(u_char) (position & 0xFF);
		else pRecompressedEvents->soft_dat[iBlob][1] = 0x00;

		pRecompressedEvents->soft_dat[iBlob][2] = (u_char)
			((pExpEvt->EvtData.hodo[iBlob].x << 5) & 0x00E0);
		if (pExpEvt->layer[iBlob] <= 5)
		    pRecompressedEvents->soft_dat[iBlob][2] =
				pRecompressedEvents->soft_dat[iBlob][2] |
				((u_char) ((position >> 8) & 0x001F));

		pRecompressedEvents->soft_dat[iBlob][3] = (u_char)
			(((pExpEvt->EvtData.hodo[iBlob].x >> 3) & 0x007F) |
			((pExpEvt->EvtData.hodo[iBlob].y << 7) & 0x0080));

		pRecompressedEvents->soft_dat[iBlob][4] = (u_char)
			((pExpEvt->EvtData.hodo[iBlob].y  >> 1) & 0xFF);
	    }
	    else {
		pRecompressedEvents->soft_dat[iBlob][0] = 0xE0;
		pRecompressedEvents->soft_dat[iBlob][1] = 0x00;
		pRecompressedEvents->soft_dat[iBlob][2] = 0x00;
		pRecompressedEvents->soft_dat[iBlob][3] = 0x00;
		pRecompressedEvents->soft_dat[iBlob][4] = 0x00;
	    }
	}

	if (pExpEvt->goodtraj > 0x00) {
	    pRecompressedEvents->coord[0][0] =
			(((short) ((pExpEvt->x1 / 0.0050) + 0.5)) & 0xFFFF);
	    pRecompressedEvents->coord[1][0] =
			(((short) ((pExpEvt->x2 / 0.0050) + 0.5)) & 0xFFFF);
	    pRecompressedEvents->coord[2][0] =
			(((short) ((pExpEvt->x3 / 0.0050) + 0.5)) & 0xFFFF);
	    pRecompressedEvents->coord[0][1] =
			(((short) ((pExpEvt->y1 / 0.0050) + 0.5)) & 0xFFFF);
	    pRecompressedEvents->coord[1][1] =
			(((short) ((pExpEvt->y2 / 0.0050) + 0.5)) & 0xFFFF);
	    pRecompressedEvents->coord[2][1] =
			(((short) ((pExpEvt->y3 / 0.0050) + 0.5)) & 0xFFFF);
	}
	else {
	    for (iXY = 0; iXY < 2; iXY++) {
		for (iLayer = 0; iLayer < 3; iLayer++) {
		    pRecompressedEvents->coord[iLayer][iXY] = 0x0000;
		}
	    }
	}

	pRecompressedEvents->blob_id =
		((((int) pExpEvt->nx1)		  & 0x0000001F) |
		((((int) pExpEvt->ny1) << 5)	  & 0x000003E0) |
		((((int) pExpEvt->nx2) << 10)	  & 0x00007C00) |
		((((int) pExpEvt->ny2) << 15)	  & 0x000F8000) |
		((((int) pExpEvt->nx3) << 20)	  & 0x01F00000) |
		((((int) pExpEvt->ny3) << 25)	  & 0x3E000000) |
		((((int) pExpEvt->goodtraj) << 30) & 0xC0000000));

	pExpEvt++;
	pRecompressedEvents++;
    }
    return pRecompressedEventsStart;
}

u_char compressNpix(long npix) {
    static u_char *compressed = NULL;
    static long *uncompressed = NULL;
    int i;
    double d;

    if (compressed == NULL) {
	compressed = (u_char *) calloc(COMPRESSED_ELEMENTS, sizeof(u_char));
	uncompressed = (long *) calloc(COMPRESSED_ELEMENTS, sizeof(long));
	for (i = 0; i < COMPRESSED_ELEMENTS; i++) {
	    compressed[i] = (u_char) i;
	    uncompressed[i] = uncompNpix(compressed[i]);
	}
    }

    d = interpol(compressed, uncompressed, npix) + 0.5;
    return (u_char) d;
}

long uncompNpix(u_char compNpix) {
    long npix = (long) compNpix;
    if (npix  >= 32) {
	long exponent = (npix >> 4) - 2L;
	long fraction = ((npix & 0x0000000F) << 1) | 0x00000021;
	npix = (fraction << exponent) - 1L;
    }
    return npix;
}

double interpol(u_char *compressed, long *uncompressed, long npix) {
    static double *slopes = NULL;
    static const int lastI = COMPRESSED_ELEMENTS - 1;
    int i;

    if (slopes == NULL) {
	slopes = (double *) calloc(COMPRESSED_ELEMENTS, sizeof(double));
	for (i = 0; i < lastI; i++) {
	    slopes[i] = ((double) (compressed[i + 1] - compressed[i])) /
			((double) (uncompressed[i + 1] - uncompressed[i]));
	}
    }

    if ((npix < uncompressed[0]) || (npix > uncompressed[lastI])) {
	fprintf(stderr, "interpol: npix = %ld out of range\n", npix);
	fprintf(stderr, "\tuncompressed[0] = %ld\n", uncompressed[0]);
	fprintf(stderr, "\tuncompressed[lastI] = %ld\n", uncompressed[lastI]);
	exit(1);
    }
    else if (npix == uncompressed[lastI]) return compressed[lastI];
    else {
	for (i = lastI; i >= 0; i--)
	    if (npix >= uncompressed[i])
		return (compressed[i] + ((npix - uncompressed[i]) * slopes[i]));
    }
    return -999.999;	/* should never get here */
}

u_char compressInt(long intensity) {
    int compressedIntensity =
		((int) ((2.0 * log(intensity + 1.0) / log(2.0)) + 0.5)) - 4;
    if (compressedIntensity < 0) return 0x00;
    else return (u_char) compressedIntensity;
}

void processSOFTException(int errorCode) {
    int verbose = VERBOSE_SOFT_EXCEPTIONS_FOR_REJECTED_EVENTS;

    if (!verbose) return;

    switch(errorCode) {
	case 421:
	    if (verbose) fprintf(stderr, "multitraj: not enough blobs\n");
	    break;

	case 422:
	    if (verbose) fprintf(stderr, "multitraj: too many blobs 1\n");
	    break;

	case 423:
	    if (verbose) fprintf(stderr, "multitraj: too many blobs 2\n");
	    break;

	case 424:
	    if (verbose) fprintf(stderr, "multitraj: too many blobs 3\n");
	    break;

	case 425:
	    if (verbose) fprintf(stderr, "multitraj: not all layers\n");
	    break;

	case 426:
	    if (verbose) fprintf(stderr, "multitraj: no X fit 1\n");
	    break;

	case 427:
	    if (verbose) fprintf(stderr, "multitraj: no X fit 2\n");
	    break;

	case 428:
	    if (verbose) fprintf(stderr, "multitraj: no Y fit 1\n");
	    break;

	case 429:
	    if (verbose) fprintf(stderr, "multitraj: no X fit 2\n");
	    break;

	case 430:
	    if (verbose) fprintf(stderr, "multitraj: no fit\n");
	    break;

	case 1000:
	    if (verbose) fprintf(stderr, "multitraj: negative detector rad\n");
	    break;

	case 1001:
	    if (verbose) fprintf(stderr, "multitraj: bad stack value\n");
	    break;

	case 1002:
	    if (verbose) fprintf(stderr, "multitraj: trajectory off E1 det\n");
	    break;
    }
}
